import asyncio
import csv
from io import TextIOWrapper
from os import path
from typing import Any, AsyncGenerator, Iterable, Union
from Bio import SeqIO

from simplest.structures.definitions import SequenceTypeDefinitions
from simplest.structures.sequences import NamedString


async def read_fasta(handle: Union[str, TextIOWrapper]) -> Iterable[NamedString]:
    fasta_sequences = asyncio.to_thread(SeqIO.parse, handle=handle, format="fasta")
    results = []
    for fasta_sequence in await fasta_sequences:
        results.append(NamedString("{0}:{1}".format(path.basename(handle.name if isinstance(handle, TextIOWrapper) else handle), fasta_sequence.id), str(fasta_sequence.seq)))
    return results

async def read_multiple_fastas(handles: Iterable[Union[str, TextIOWrapper]]) -> AsyncGenerator[Iterable[NamedString], Any]:
    tasks = []
    for handle in handles:
        tasks.append(read_fasta(handle))
    for task in asyncio.as_completed(tasks):
        yield await task

def read_sequence_type_definitions(handle: Union[str, TextIOWrapper], sequence_type_header="ST") -> SequenceTypeDefinitions:
    with open(handle) if isinstance(handle, str) else handle as st_def_files:
        dict_reader = csv.DictReader(st_def_files)
        st_definitions: list[tuple[dict[str, str], str]] = list()
        loci = list()
        for row in dict_reader:
            locus_alleles = dict()
            for locus in row.keys():
                if locus == sequence_type_header:
                    continue
                locus_alleles[locus] = str(row[locus])
            st_definitions.append((locus_alleles, str(row[sequence_type_header])))
            if len(loci) == 0:
                loci.extend(locus_alleles.keys())
        return SequenceTypeDefinitions(loci, st_definitions)