import asyncio
from concurrent.futures import Future, ThreadPoolExecutor
from contextlib import AbstractAsyncContextManager, AbstractContextManager
from os import path
import os
from types import CoroutineType
from typing import Any, Iterable, Mapping, Set, Union
from Bio.Align import PairwiseAligner
from queue import Queue

from simplest.io.readers import read_fasta
from simplest.structures.alignment import AlignmentStats, PairwiseAlignment
from simplest.structures.definitions import SequenceTypeDefinitions
from simplest.structures.mlst import Allele, MLSTProfile
from simplest.structures.sequences import NamedString

from logging import getLogger

logger = getLogger(__name__)

class SequenceTyper(AbstractContextManager):
    def __enter__(self):
        return self

    def __init__(self, sequence_type_definitions: SequenceTypeDefinitions, allele_fasta_definitions_dir: str, max_threads = 4):
        self._st_definitions = sequence_type_definitions
        self._allele_dir = allele_fasta_definitions_dir
        self._max_threads = max_threads
        self._thread_pool = ThreadPoolExecutor(max_workers=self._max_threads, thread_name_prefix="sequence-typer")

    def local_pairwise_alignment_with_metadata(self, reference: str, query: str, **data):
        aligner = PairwiseAligner("blastn")
        aligner.mode = "local"
        alignment_candidate_result = aligner.align(query.upper(), reference.upper())
        top_alignment = sorted(alignment_candidate_result)[0] # TODO might need sorting...?
        top_alignment_stats = top_alignment.counts()
        top_alignment_gaps = top_alignment_stats.gaps
        top_alignment_identities = top_alignment_stats.identities
        top_alignment_mismatches = top_alignment_stats.mismatches
        top_alignment_score = top_alignment.score # type: ignore
        alignment_data = PairwiseAlignment(
            top_alignment.sequences[0],
            top_alignment.sequences[1],
            tuple(top_alignment.indices[0]),
            tuple(top_alignment.indices[1]),
            AlignmentStats(
                percent_identity=top_alignment_identities/top_alignment.length,
                mismatches=top_alignment_mismatches,
                gaps=top_alignment_gaps,
                match_metric=top_alignment_score
            )), data
        return alignment_data

    async def calculate_allele_of_locus(self, locus: str, query: NamedString):
        possible_alleles_paths = [path.join(self._allele_dir, locus, item) for item in os.listdir(path.join(self._allele_dir, locus))]
        alignments = list()
        for possible_allele in possible_alleles_paths:
            fastas = list(await read_fasta(possible_allele))
            target_fasta = fastas[0] # TODO improve handling of multi-sequence fasta
            alignments.append(asyncio.wrap_future(self._thread_pool.submit(
                self.local_pairwise_alignment_with_metadata,
                query.sequence,
                target_fasta.sequence,
                query_name=query.name,
                allele_name=path.splitext(path.basename(possible_allele))[0]
            )))
        
        alignment_results: list[tuple[PairwiseAlignment, Mapping[str, Any]]] = list()
        for alignment_candidate in asyncio.as_completed(alignments):
            alignment_result = await alignment_candidate
            alignment_results.append(alignment_result)

        alignment_results = sorted(alignment_results, key=lambda item: item[0].alignment_stats.match_metric, reverse=True)
        best_alignment, alignment_data = alignment_results[0]
        allele_name = alignment_data["allele_name"]
        return Allele(locus, allele_name, best_alignment.alignment_stats if best_alignment.alignment_stats.percent_identity != 1 else None)

    async def calculate_st_from_sequence_paths(self, sequences: Mapping[str, str]):
        alleles = list()
        for locus, path in sequences.items():
            if path is None:
                alleles.append(Allele(locus, "missing", None))
                continue
            query_sequences = list(await read_fasta(path))
            query = query_sequences[0]
            alleles.append(await self.calculate_allele_of_locus(locus, query))
        return MLSTProfile(alleles, self._st_definitions.get_sequence_type_from_alleles(alleles), "none")

    def shutdown(self):
        self._thread_pool.shutdown()

    def __exit__(self, exc_type, exc_value, traceback):
        self.shutdown()

class DirectoryQuerySequenceManager:
    def __init__(self, root: str, sequence_type_definitions: SequenceTypeDefinitions):
        self._root = root
        self._sequence_type_Definitions = sequence_type_definitions

    def check(self):
        available_locis = set(os.listdir(self._root))
        for locus in self._sequence_type_Definitions.get_ordered_locis():
            if locus not in available_locis:
                return False
        return True
    
    async def group_query_sequences(self):
        all_sample_ids = set()
        extensions = set()
        for locus in self._sequence_type_Definitions.get_ordered_locis():
            directory = path.join(self._root, locus)
            for item in os.listdir(directory):
                name, ext = path.splitext(path.basename(item))
                all_sample_ids.add(name)
                extensions.add(ext)
        for sample_id in all_sample_ids:
            sample_ordered_query = dict()
            for locus in self._sequence_type_Definitions.get_ordered_locis():
                found = False
                for extension in extensions:
                    target_sample_path = path.join(self._root, locus, sample_id + extension)
                    if path.isfile(target_sample_path):
                        sample_ordered_query[locus] = target_sample_path
                        found = True
                        break
                if not found:
                    sample_ordered_query[locus] = None
            yield sample_ordered_query, sample_id