import argparse
import asyncio
import os

from simplest.io.readers import read_sequence_type_definitions
from simplest.io.writers import write_mlst_profiles_as_csv
from simplest.structures.definitions import SequenceTypeDefinitions
from simplest.structures.mlst import NamedMLSTProfile
from simplest.typing.profilers import DirectoryQuerySequenceManager, SequenceTyper
from logging import getLogger

async def run():
    logger = getLogger(__name__)

    parser = argparse.ArgumentParser("simplest")
    parser.add_argument(
        "ST",
        type=str,
        help="The spreadsheet (in CSV format) that contains the sequence type definitions."
    )
    parser.add_argument(
        "alleles",
        type=str,
        help="The directory containing all the alleles and their sequence definitions."
    )
    parser.add_argument(
        "FASTA",
        type=str,
        help="FASTA consensus files to type."
    )
    parser.add_argument(
        "out",
        type=str,
        help="Output file."
    )
    parser.add_argument(
        "-sth", "--sequence-type-header",
        type=str,
        default="ST",
        required=False,
        help="The column name/header representing the column of sequence types. Defaults to \"ST\""
    )
    args = parser.parse_args()
    st_definitions = read_sequence_type_definitions(args.ST, args.sequence_type_header)
    with SequenceTyper(st_definitions, allele_fasta_definitions_dir=args.alleles) as typer:
        query_manager = DirectoryQuerySequenceManager(args.FASTA, st_definitions)
        results = []
        async for query_seq, sample_id in query_manager.group_query_sequences():
            print(f"Queueing sample set: {query_seq}")
            results.append(NamedMLSTProfile(sample_id, await typer.calculate_st_from_sequence_paths(query_seq)))
        failed = await write_mlst_profiles_as_csv(results, args.out, st_definitions.get_ordered_locis())
        if len(failed) > 0:
            print(failed)

def main():
    asyncio.run(run())

if __name__ == "__main__":
    main()